#!/usr/bin/env node
'use strict';

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _cpr = require('cpr');

var _cpr2 = _interopRequireDefault(_cpr);

var _minimist = require('minimist');

var _minimist2 = _interopRequireDefault(_minimist);

var _generate2 = require('../generate');

var _generate3 = _interopRequireDefault(_generate2);

var _child_process = require('child_process');

var _child_process2 = _interopRequireDefault(_child_process);

var _escapeStringRegexp = require('escape-string-regexp');

var _escapeStringRegexp2 = _interopRequireDefault(_escapeStringRegexp);

var _md = require('md5');

var _md2 = _interopRequireDefault(_md);

var _chalk = require('chalk');

var _chalk2 = _interopRequireDefault(_chalk);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var argv = (0, _minimist2.default)(process.argv.slice(2), {
  default: {
    'force-update': false
  }
});

/* eslint-disable no-console */

var commands = argv._;
var SKEL_DIR = _path2.default.join(__dirname, '..', 'skel');
var CHECKSUM_FILE = _path2.default.join(process.cwd(), '.create-graphql-server.checksums');

function usage() {
  console.log('Generate a GraphQL Server environment, or add/update/remove GraphQL data types and generate code files and schemas.');
  console.log('');
  console.log('Usage:', _chalk2.default.bold.yellow('create-graphql-server <command> <argument> <option>, where command is:'));
  console.log('');
  console.log(' - create-graphql-server init project-dir');
  console.log(' - create-graphql-server add-type path/to/type.graphql');
  console.log(' - create-graphql-server add-type path');
  console.log(' - create-graphql-server remove-type path/to/type.graphql');
  console.log(' - create-graphql-server remove-type path');
  console.log(' - create-graphql-server remove-type typename');
  console.log('');
  console.log(_chalk2.default.yellow(' Commands:'));
  console.log('');
  console.log(' init           initializes and generates a new graphql-server project environment');
  console.log(' add-type       adds your new Graphql data type to the graphql-server');
  console.log(' remove-type    removes a Graphql data type from the graphql-server (does not delete data in db)');
  console.log('');
  console.log(_chalk2.default.yellow(' Argument:'));
  console.log('');
  console.log(' file-path      path of your <type>.graphql file e.g.: ./input/User.graphql');
  console.log('');
  console.log(_chalk2.default.yellow(' Options:'));
  console.log('');
  console.log(' --force-update forces updates, overwriting code changes by any user since last run');
  console.log('');
  process.exit(1);
}

function adjustTypeName(typeName) {
  return typeName.charAt(0).toUpperCase() + typeName.slice(1).toLowerCase();
}

function getFileUpdateList(inputSchemaFile, mode) {
  var inputSchemaStr = '';
  // in add mode or if a graphql path/file name was provided, the input file must be there,
  if (mode === 'add' || inputSchemaFile.includes('.graphql') || inputSchemaFile.includes('/')) {
    if (!_fs2.default.existsSync(inputSchemaFile)) {
      console.error(_chalk2.default.bold.red('Error: Cannot read file', inputSchemaFile));
      console.log('');
      process.exit(1);
    }
    // get information about the entity <type> out of its input file
    inputSchemaStr = _fs2.default.readFileSync(inputSchemaFile, 'utf8');
  } else {
    // expecting just the type name without path/file name
    // provide a dummy type file just with the type name
    inputSchemaStr = 'type ' + adjustTypeName(inputSchemaFile) + ' {}';
  }

  // generate code therefore in memory first

  var _generate = (0, _generate3.default)(inputSchemaStr),
      typeName = _generate.typeName,
      TypeName = _generate.TypeName,
      outputSchemaStr = _generate.outputSchemaStr,
      resolversStr = _generate.resolversStr,
      modelStr = _generate.modelStr;

  // do validation checks
  // shouldn't be necessary, but...


  if (typeName === '' || TypeName === '') {
    console.error('Error: No valid <Type> found.');
    process.exit(0);
  }
  // there should be some generated code, not adding missing or empty code files
  if (!outputSchemaStr || !resolversStr || !modelStr || outputSchemaStr === '' || resolversStr === '' || modelStr === '') {
    console.error('Error: Error while generating target Code.');
    process.exit(0);
  }

  // Bill-of-material for the changes to be applied
  // Add further enhancements to this array
  //
  // @Records:
  // One record represents one file <type> with its path, code and reference
  //
  // @Fields:
  // typePath: path and file name for the <type> file
  // typeString: generated code for the <type> file
  // indexPath: path and file name for the index file, which has to reference the <type> file
  // indexPattern: code pattern to be added to the index file, for referencing this <type> file
  return [{
    typePath: _path2.default.join('schema', TypeName + '.graphql'),
    typeString: outputSchemaStr,
    indexPath: _path2.default.join('schema', 'index.js'),
    indexPattern: '\ntypeDefs.push(requireGraphQL(\'./' + TypeName + '.graphql\'));\n'
  }, {
    typePath: _path2.default.join('resolvers', TypeName + '.js'),
    typeString: resolversStr,
    indexPath: _path2.default.join('resolvers', 'index.js'),
    indexPattern: '\nimport ' + typeName + 'Resolvers from \'./' + TypeName + '\';\n' + ('merge(resolvers, ' + typeName + 'Resolvers);\n')
  }, {
    typePath: _path2.default.join('model', TypeName + '.js'),
    typeString: modelStr,
    indexPath: _path2.default.join('model', 'index.js'),
    indexPattern: '\nimport ' + TypeName + ' from \'./' + TypeName + '\';\n' + ('models.' + TypeName + ' = ' + TypeName + ';\n')
  }];
}

function getEscapeRegex(pattern) {
  // Prepare Regular Expression to find and replace code patterns
  // Special characters in the code string have to be escaped, to get the regex working properly
  // First with common escape patterns by the included npm module,
  // Additionally escape ./ and \n, which weren't escaped by the npm module unfortunately
  var re = new RegExp((0, _escapeStringRegexp2.default)(pattern) // use common escape patterns
  .replace(/\.\//g, '.\\/') // escape: ./  to   \.\/ (which was in path)
  .replace(/\n/g, '\\n'), // escape: \n  to:  \\n
  'gmi');
  return re;
}

function getFileHash(fileContent) {
  return (0, _md2.default)(fileContent);
}

function readFileHashFile() {
  // read file with all file hashes from last add-type/update-type/remove-type runs
  var checksums = {};
  if (_fs2.default.existsSync(CHECKSUM_FILE)) {
    checksums = JSON.parse(_fs2.default.readFileSync(CHECKSUM_FILE, 'utf8'));
  }
  return checksums;
}

function removeFileHash(file) {
  // MD5_file_Hash_JSON = [ { 'User.graphql': 'md5abscdsadfasf234322'} ]
  var checksums = [];
  // Existing Hash File: Read and update file, remove the entry with <file>
  if (file && file !== '' && _fs2.default.existsSync(CHECKSUM_FILE)) {
    checksums = JSON.parse(_fs2.default.readFileSync(CHECKSUM_FILE, 'utf8'));
    delete checksums[file];
    _fs2.default.writeFileSync(CHECKSUM_FILE, JSON.stringify(checksums));
  }
}

function writeFileHashFile(file, content) {
  // MD5_file_Hash_JSON = { 'User.graphql': 'md5abscdsadfasf234322'}
  var checksums = {};
  // Existing Hash File: Read and update file
  if (file && file !== '' && content && content !== '' && _fs2.default.existsSync(CHECKSUM_FILE)) {
    checksums = JSON.parse(_fs2.default.readFileSync(CHECKSUM_FILE, 'utf8'));
    checksums[file] = getFileHash(content);
    _fs2.default.writeFileSync(CHECKSUM_FILE, JSON.stringify(checksums));
    // Not yet existing Hash File: create file
  } else if (file && file !== '' && content && content !== '') {
    checksums[file] = getFileHash(content);
    _fs2.default.writeFileSync(CHECKSUM_FILE, JSON.stringify(checksums));
  }
}

function hasFileChanged(file, checksums) {
  // file has changed => true,
  // file not changed => false
  // file not found   => false
  // If file is provided and existing,
  // and also an old entry in the file is existing,
  // then compare old and new hash
  if (file && file !== '' && checksums[file] && _fs2.default.existsSync(file)) {
    var oldFileHash = checksums[file];
    var newFileHash = getFileHash(_fs2.default.readFileSync(file, 'utf8'));
    if (oldFileHash !== newFileHash) {
      return true;
    }
  }
  return false;
}

function checkForFileChanges(fileUpdateList, forceUpdate) {
  // Get all old file hashes first
  var checksums = readFileHashFile();
  var anyFileHasChanged = false;
  var thisFileHasChanged = false;
  if (checksums) {
    fileUpdateList.map(function (file) {
      // Check if <type> file has changed
      thisFileHasChanged = hasFileChanged(file.typePath, checksums);
      if (thisFileHasChanged) {
        console.log(_chalk2.default.bold.red('Caution: File has changed:', file.typePath));
        anyFileHasChanged = anyFileHasChanged || thisFileHasChanged;
      }
      // Check if index file has changed
      thisFileHasChanged = hasFileChanged(file.indexPath, checksums);
      if (thisFileHasChanged) {
        console.log(_chalk2.default.bold.red('Caution: File has changed:', file.indexPath));
        anyFileHasChanged = anyFileHasChanged || thisFileHasChanged;
      }
      return file;
    });
    // if any file has changed, stop
    if (anyFileHasChanged && !forceUpdate) {
      console.error(_chalk2.default.bold.red('Stopping without change for safety reasons.'));
      console.error(_chalk2.default.green('Please check you are not overwriting your changes.'));
      console.error(_chalk2.default.green('You can still overwrite by using the --force-update option.'));
      console.error(' ');
      process.exit(0);
      // if any file has changed, but user chose force-update, then continue with changing files...
    } else if (anyFileHasChanged && forceUpdate) {
      console.log(_chalk2.default.bold.yellow('Overwriting files --force-update option was chosen...'));
    }
  }
}

function checkFileContainsPattern(file, pattern) {
  // only checks, if the file contains a search pattern
  if (pattern && file && _fs2.default.existsSync(file)) {
    var inputFile = _fs2.default.readFileSync(file, 'utf8');
    var re = getEscapeRegex(pattern);
    var found = inputFile.match(re);
    if (found === null) {
      return false; // pattern not found
    }
    return true; // pattern found
  }
  console.error('Error: checkFileContainsPattern() requires an existing file:', file, ' and a pattern:', pattern);
  process.exit(0);
  return false;
}

function removePatternFromFile(file, pattern) {
  // Remove <type> related code patterns from file
  // which were added earlier by "add-type" to e.g. index.js file
  if (file && pattern && _fs2.default.existsSync(file)) {
    var inputFile = _fs2.default.readFileSync(file, 'utf8');
    var re = getEscapeRegex(pattern);
    var targetFile = inputFile.replace(re, '');
    _fs2.default.writeFileSync(file, targetFile);
    writeFileHashFile(file, targetFile);
    console.log(_chalk2.default.green('File updated:', file));
  }
}

function removePatternFromFiles(fileUpdateList) {
  fileUpdateList.map(function (file) {
    removePatternFromFile(file.indexPath, file.indexPattern);
    return file;
  });
}

function removeFile(file) {
  // deletes the file on existence
  if (file && file !== '' && _fs2.default.existsSync(file)) {
    _fs2.default.unlinkSync(file);
    removeFileHash(file);
    console.log(_chalk2.default.green('File removed:', file));
  }
}

function removeTypeFiles(fileUpdateList) {
  fileUpdateList.map(function (file) {
    removeFile(file.typePath);
    return file;
  });
}

function createTypeFile(file, content) {
  // creates the <type> file with its generated code
  if (file && content && file !== '') {
    _fs2.default.writeFileSync(file, content);
    writeFileHashFile(file, content);
    console.log(_chalk2.default.green('File generated:', file));
  }
}

function createTypeFiles(fileUpdateList) {
  fileUpdateList.map(function (file) {
    createTypeFile(file.typePath, file.typeString);
    return file;
  });
}

function addPatternToFile(file, pattern) {
  // appends the code pattern to the index file, for referencing the <type> file
  if (checkFileContainsPattern(file, pattern)) {
    // file contains pattern already, do nothing
    console.log(_chalk2.default.green('Skipping file:', file, 'code pattern already in file.'));
  } else if (_fs2.default.existsSync(file)) {
    _fs2.default.appendFileSync(file, pattern);
    var content = _fs2.default.readFileSync(file, 'utf8');
    writeFileHashFile(file, content);
    console.log(_chalk2.default.green('File enhanced: ', file, 'with required reference.'));
  } else {
    console.error(_chalk2.default.bold.red('ERROR: File to be enhanced: ', file, 'with required reference, does not exist.'));
  }
}

function addPatternToFiles(fileUpdateList) {
  fileUpdateList.map(function (file) {
    addPatternToFile(file.indexPath, file.indexPattern);
    return file;
  });
}

function getFilesRecursively(folder, filetype) {
  // getting all files of a path and of a specific filetype recursively
  var list = [],
      stats = void 0,
      files = _fs2.default.readdirSync(folder);

  files.forEach(function (file) {
    stats = _fs2.default.lstatSync(_path2.default.join(folder, file));
    if (stats.isDirectory()) {
      list = list.concat(getFilesRecursively(_path2.default.join(folder, file), filetype));
    } else if (file.includes(filetype)) {
      console.log('found:', _path2.default.join(folder, file));
      list.push(_path2.default.join(folder, file));
    }
  });

  return list;
}

// MAIN FUNCTIONS

function addType(inputSchemaFile, options) {
  // generates a new data type with all its files and <type> references
  console.log(_chalk2.default.bold.blue('Running add-type'));
  var fileUpdateList = getFileUpdateList(inputSchemaFile, 'add');
  checkForFileChanges(fileUpdateList, options['force-update']);
  createTypeFiles(fileUpdateList);
  addPatternToFiles(fileUpdateList);
  console.log(_chalk2.default.bold.blue('Finished successfully add-type'));
  console.log('');
}

function removeType(inputSchemaFile, options) {
  // removes the generated files and references only
  console.log(_chalk2.default.bold.blue('Running remove-type'));
  var fileUpdateList = getFileUpdateList(inputSchemaFile, 'remove');
  checkForFileChanges(fileUpdateList, options['force-update']);
  removePatternFromFiles(fileUpdateList);
  removeTypeFiles(fileUpdateList);
  console.log(_chalk2.default.bold.blue('Finished successfully remove-type'));
  console.log('');
}

// CHECK AND REACT ON USER INPUTS
console.log('');
console.log(_chalk2.default.bold.blue('CREATE-GRAPHQL-SERVER'));
console.log('');

if (commands[0] === 'init') {
  (function () {
    var projectDir = commands[1];
    if (!projectDir) {
      usage();
    }
    // JWT key for encrypting tokens
    var key = process.env.JWT_KEY || Math.random().toString();
    // TODO - nice things like checking the directory doesn't exist
    (0, _cpr2.default)(SKEL_DIR, projectDir, { confirm: true, overwrite: true }, function () {
      var command = 'find  * -type f  -exec sed -i \'\' -e \'s/~name~/' + projectDir + '/\' -e \'s/~key~/' + key + '/\' {} \\;';
      _child_process2.default.exec(command, { cwd: './' + projectDir }, function () {
        console.log('Created project in ' + projectDir);
        process.exit(0);
      });
    });
  })();
} else if (commands[0] === 'add-type') {
  var inputSchemaFile = commands[1];
  if (!inputSchemaFile) {
    usage();
  }
  if (_fs2.default.existsSync(inputSchemaFile) && _fs2.default.lstatSync(inputSchemaFile).isDirectory()) {
    //directory name entered
    var files = getFilesRecursively(inputSchemaFile, '.graphql');
    files.forEach(function (file) {
      addType(file, argv);
    });
  } else {
    // single file entered
    addType(inputSchemaFile, argv);
  }
  process.exit(0);
} else if (commands[0] === 'remove-type') {
  var _inputSchemaFile = commands[1];
  if (!_inputSchemaFile) {
    usage();
  }
  if (_fs2.default.existsSync(_inputSchemaFile) && _fs2.default.lstatSync(_inputSchemaFile).isDirectory()) {
    // directory name entered
    var _files = getFilesRecursively(_inputSchemaFile, '.graphql');
    _files.forEach(function (file) {
      removeType(file, argv);
    });
  } else {
    // single file or type
    removeType(_inputSchemaFile, argv);
  }
  process.exit(0);
} else {
  usage();
}